@@ -265,11 +265,20 @@ class Script(db.Model):
         output_variables: Dict[str, str] = self.get_variables()
         # print(output_variables)
         for key, value in kwargs.items():
-            if type(value) is str and value in output_variables:
-                var_type = output_variables[value]
-                kwargs[key] = {value: var_type}
-            if isinstance(value, str) and value.startswith("#"):
-                kwargs[key] = f"#{self.validate_function_name(value[1:])}"
+            if isinstance(value, str):
+                if value in output_variables:
+                    var_type = output_variables[value]
+                    kwargs[key] = {value: var_type}
+                elif value.startswith("#"):
+                    kwargs[key] = f"#{self.validate_function_name(value[1:])}"
+                else:
+                    # attempt to convert to numerical or bool value for args with no type hint
+                    try:
+                        converted = ast.literal_eval(value)
+                        if isinstance(converted, (int, float, bool)):
+                            kwargs[key] = converted
+                    except (ValueError, SyntaxError):
+                        pass
         return kwargs
 
     def add_logic_action(self, logic_type: str, statement, insert_position=None):
@@ -309,6 +318,12 @@ class Script(db.Model):
                     {"id": current_len + 2, "instrument": 'repeat', "action": 'endrepeat',
                      "args": {}, "return": '', "uuid": uid},
                 ],
+            "pause":
+                [
+                    {"id": current_len + 1, "instrument": 'pause', "action": "pause",
+                     "args": {"statement": 1 if statement == '' else statement}, "return": '', "uuid": uid,
+                     "arg_types": {"statement": "str"}}
+                ],
         }
         action_list = logic_dict[logic_type]
         self.currently_editing_script.extend(action_list)
@@ -434,6 +449,9 @@ class Script(db.Model):
         Compile the current script to a Python file.
         :return: String to write to a Python file.
         """
+        self.needs_call_human = False
+        self.blocks_included = False
+
         self.sort_actions()
         run_name = self.name if self.name else "untitled"
         run_name = self.validate_function_name(run_name)
@@ -465,7 +483,7 @@ class Script(db.Model):
         """
         configure, config_type = self.config(stype)
 
-        configure = [param + f":{param_type}" if not param_type == "any" else "" for param, param_type in
+        configure = [param + f":{param_type}" if not param_type == "any" else param for param, param_type in
                      config_type.items()]
 
         script_type = f"_{stype}" if stype != "script" else ""
@@ -515,6 +533,9 @@ class Script(db.Model):
             return f"{self.indent(indent_unit)}time.sleep({statement})", indent_unit
         elif instrument == 'repeat':
             return self._process_repeat(indent_unit, action_name, statement, next_action)
+        elif instrument == 'pause':
+            self.needs_call_human = True
+            return f"{self.indent(indent_unit)}pause('{statement}')", indent_unit
         #todo
         # elif instrument == 'registered_workflows':
         #     return inspect.getsource(my_function)
@@ -583,14 +604,18 @@ class Script(db.Model):
         """
         Process actions related to instruments.
         """
+        function_call = f"{instrument}.{action}"
+        if instrument.startswith("blocks"):
+            self.blocks_included = True
+            function_call = action
 
-        if isinstance(args, dict):
+        if isinstance(args, dict) and args != {}:
             args_str = self._process_dict_args(args)
-            single_line = f"{instrument}.{action}(**{args_str})"
+            single_line = f"{function_call}(**{args_str})"
         elif isinstance(args, str):
-            single_line = f"{instrument}.{action} = {args}"
+            single_line = f"{function_call} = {args}"
         else:
-            single_line = f"{instrument}.{action}()"
+            single_line = f"{function_call}()"
 
         if save_data:
             save_data += " = "
@@ -631,7 +656,7 @@ class Script(db.Model):
         """
         return arg in self.script_dict and self.script_dict[arg].get("arg_types") == "variable"
 
-    def _write_to_file(self, script_path, run_name, exec_string):
+    def _write_to_file(self, script_path, run_name, exec_string, call_human=False):
         """
         Write the compiled script to a file.
         """
@@ -641,10 +666,30 @@ class Script(db.Model):
             else:
                 s.write("deck = None")
             s.write("\nimport time")
+            if self.blocks_included:
+                s.write(f"\n{self._create_block_import()}")
+            if self.needs_call_human:
+                s.write("""\n\ndef pause(reason="Manual intervention required"):\n\tprint(f"\\nHUMAN INTERVENTION REQUIRED: {reason}")\n\tinput("Press Enter to continue...\\n")""")
+
             for i in exec_string.values():
                 s.write(f"\n\n\n{i}")
 
+    def _create_block_import(self):
+        imports = {}
+        from ivoryos.utils.decorators import BUILDING_BLOCKS
+        for category, methods in BUILDING_BLOCKS.items():
+            for method_name, meta in methods.items():
+                func = meta["func"]
+                module = meta["path"]
+                name = func.__name__
+                imports.setdefault(module, set()).add(name)
+        lines = []
+        for module, funcs in imports.items():
+            lines.append(f"from {module} import {', '.join(sorted(funcs))}")
+        return "\n".join(lines)
+
 class WorkflowRun(db.Model):
+    """Represents the entire experiment"""
     __tablename__ = 'workflow_runs'
 
     id = db.Column(db.Integer, primary_key=True)
@@ -653,30 +698,65 @@ class WorkflowRun(db.Model):
     start_time = db.Column(db.DateTime, default=datetime.now())
     end_time = db.Column(db.DateTime)
     data_path = db.Column(db.String(256))
-    steps = db.relationship(
-        'WorkflowStep',
-        backref='workflow_runs',
+    repeat_mode = db.Column(db.String(64), default="none")  # static_repeat, sweep, optimizer
+
+    # A run contains multiple iterations
+    phases = db.relationship(
+        'WorkflowPhase',
+        backref='workflow_runs', # Clearer back-reference name
         cascade='all, delete-orphan',
-        passive_deletes=True
+        lazy='dynamic' # Good for handling many iterations
     )
     def as_dict(self):
         dict = self.__dict__
         dict.pop('_sa_instance_state', None)
         return dict
 
+class WorkflowPhase(db.Model):
+    """Represents a single function call within a WorkflowRun."""
+    __tablename__ = 'workflow_phases'
+
+    id = db.Column(db.Integer, primary_key=True)
+    # Foreign key to link this iteration to its parent run
+    run_id = db.Column(db.Integer, db.ForeignKey('workflow_runs.id', ondelete='CASCADE'), nullable=False)
+
+    # NEW: Store iteration-specific parameters here
+    name = db.Column(db.String(64), nullable=False)  # 'prep', 'main', 'cleanup'
+    repeat_index = db.Column(db.Integer, default=0)
+
+    parameters = db.Column(JSONType)  # Use db.JSON for general support
+    outputs = db.Column(JSONType)
+    start_time = db.Column(db.DateTime, default=datetime.now)
+    end_time = db.Column(db.DateTime)
+
+    # An iteration contains multiple steps
+    steps = db.relationship(
+        'WorkflowStep',
+        backref='workflow_phases',  # Clearer back-reference name
+        cascade='all, delete-orphan'
+    )
+
+    def as_dict(self):
+        dict = self.__dict__.copy()
+        dict.pop('_sa_instance_state', None)
+        return dict
+
 class WorkflowStep(db.Model):
     __tablename__ = 'workflow_steps'
 
     id = db.Column(db.Integer, primary_key=True)
-    workflow_id = db.Column(db.Integer, db.ForeignKey('workflow_runs.id', ondelete='CASCADE'), nullable=False)
+    # workflow_id = db.Column(db.Integer, db.ForeignKey('workflow_runs.id', ondelete='CASCADE'), nullable=True)
+    phase_id = db.Column(db.Integer, db.ForeignKey('workflow_phases.id', ondelete='CASCADE'), nullable=True)
 
-    phase = db.Column(db.String(64), nullable=False)  # 'prep', 'main', 'cleanup'
-    repeat_index = db.Column(db.Integer, default=0)   # Only applies to 'main' phase
+    # phase = db.Column(db.String(64), nullable=False)  # 'prep', 'main', 'cleanup'
+    # repeat_index = db.Column(db.Integer, default=0)   # Only applies to 'main' phase
     step_index = db.Column(db.Integer, default=0)
     method_name = db.Column(db.String(128), nullable=False)
     start_time = db.Column(db.DateTime)
     end_time = db.Column(db.DateTime)
     run_error = db.Column(db.Boolean, default=False)
+    output = db.Column(JSONType, default={})
+    # Using as_dict method from ModelBase
 
     def as_dict(self):
         dict = self.__dict__.copy()
@@ -693,7 +773,7 @@ class SingleStep(db.Model):
     start_time = db.Column(db.DateTime)
     end_time = db.Column(db.DateTime)
     run_error = db.Column(db.String(128))
-    output = db.Column(JSONType)
+    output = db.Column(JSONType, nullable=True)
 
     def as_dict(self):
         dict = self.__dict__.copy()
