@@ -1,5 +1,10 @@
 from enum import Enum
-from typing import get_origin, get_args, Union, Any
+from typing import Union, Any
+try:
+    from typing import get_origin, get_args
+except ImportError:
+    # For Python versions = 3.7, use typing_extensions
+    from typing_extensions import get_origin, get_args
 
 from wtforms.fields.choices import SelectField
 from wtforms.fields.core import Field
@@ -205,20 +210,15 @@ class FlexibleEnumField(StringField):
             if key in self.choices:
                 # Convert the string key to Enum instance
                 self.data = self.enum_class[key].value
-            elif self.data.startswith("#"):
+            elif key.startswith("#"):
                 if not self.script.editing_type == "script":
                     raise ValueError(self.gettext("Variable is not supported in prep/cleanup"))
-                self.data = self.data
+                self.data = key
             else:
                 raise ValidationError(
                     f"Invalid choice: '{key}'. Must match one of {list(self.enum_class.__members__.keys())}")
 
 
-def format_name(name):
-    """Converts 'example_name' to 'Example Name'."""
-    name = name.split(".")[-1]
-    text = ' '.join(word for word in name.split('_'))
-    return text.capitalize()
 
 def parse_annotation(annotation):
     """
@@ -264,7 +264,7 @@ def create_form_for_method(method, autofill, script=None, design=True):
     for param in sig.parameters.values():
         if param.name == 'self':
             continue
-        formatted_param_name = format_name(param.name)
+        # formatted_param_name = format_name(param.name)
 
         default_value = None
         if autofill:
@@ -277,7 +277,7 @@ def create_form_for_method(method, autofill, script=None, design=True):
                     default_value = param.default
 
         field_kwargs = {
-            "label": formatted_param_name,
+            "label": param.name,
             "default": default_value,
             "validators": [InputRequired()] if param.default is param.empty else [Optional()],
             **({"script": script} if (autofill or design) else {})
@@ -286,7 +286,9 @@ def create_form_for_method(method, autofill, script=None, design=True):
             # enum_class = [(e.name, e.value) for e in param.annotation]
             field_class = FlexibleEnumField
             placeholder_text = f"Choose or type a value for {param.annotation.__name__} (start with # for custom)"
+
             extra_kwargs = {"choices": param.annotation}
+
         else:
             # print(param.annotation)
             annotation, optional = parse_annotation(param.annotation)
@@ -299,6 +301,11 @@ def create_form_for_method(method, autofill, script=None, design=True):
             if optional:
                 field_kwargs["filters"] = [lambda x: x if x != '' else None]
 
+            if annotation is bool:
+                # Boolean fields should not use InputRequired
+                field_kwargs["validators"] = []  # or [Optional()]
+            else:
+                field_kwargs["validators"] = [InputRequired()] if param.default is param.empty else [Optional()]
 
         render_kwargs = {"placeholder": placeholder_text}
 
@@ -396,12 +403,12 @@ def create_form_from_action(action: dict, script=None, design=True):
     }
 
     for name, param_type in arg_types.items():
-        formatted_param_name = format_name(name)
+        # formatted_param_name = format_name(name)
         value = args.get(name, "")
         if type(value) is dict:
             value = next(iter(value))
         field_kwargs = {
-            "label": formatted_param_name,
+            "label": name,
             "default": f'{value}',
             "validators": [InputRequired()],
             **({"script": script})
@@ -424,7 +431,7 @@ def create_form_from_action(action: dict, script=None, design=True):
 
 def create_all_builtin_forms(script):
     all_builtin_forms = {}
-    for logic_name in ['if', 'while', 'variable', 'wait', 'repeat']:
+    for logic_name in ['if', 'while', 'variable', 'wait', 'repeat', 'pause']:
         # signature = info.get('signature', {})
         form_class = create_builtin_form(logic_name, script)
         all_builtin_forms[logic_name] = form_class()
@@ -439,7 +446,8 @@ def create_builtin_form(logic_type, script):
 
     placeholder_text = {
         'wait': 'Enter second',
-        'repeat': 'Enter an integer'
+        'repeat': 'Enter an integer',
+        'pause': 'Human Intervention Message'
     }.get(logic_type, 'Enter statement')
     description_text = {
         'variable': 'Your variable can be numbers, boolean (True or False) or text ("text")',
@@ -531,6 +539,7 @@ def _action_button(action: dict, variables: dict):
         "repeat": "background-color: lightsteelblue",
         "if": "background-color: salmon",
         "while": "background-color: salmon",
+        "pause": "background-color: goldenrod",
     }.get(action['instrument'], "")
 
     if action['instrument'] in ['if', 'while', 'repeat']:
