@@ -6,14 +6,28 @@ import time
 from datetime import datetime
 
 from ivoryos.utils import utils, bo_campaign
-from ivoryos.utils.db_models import Script, WorkflowRun, WorkflowStep, db, SingleStep
+from ivoryos.utils.db_models import Script, WorkflowRun, WorkflowStep, db, WorkflowPhase
 from ivoryos.utils.global_config import GlobalConfig
+from ivoryos.utils.decorators import BUILDING_BLOCKS
 
 global_config = GlobalConfig()
 global deck
 deck = None
 # global deck, registered_workflows
 # deck, registered_workflows = None, None
+class HumanInterventionRequired(Exception):
+    pass
+
+def pause(reason="Human intervention required"):
+    handlers = global_config.notification_handlers
+    if handlers:
+        for handler in handlers:
+            try:
+                handler(reason)
+            except Exception as e:
+                print(f"[notify] handler {handler} failed: {e}")
+    # raise error to pause workflow in gui
+    raise HumanInterventionRequired(reason)
 
 class ScriptRunner:
     def __init__(self, globals_dict=None):
@@ -62,11 +76,11 @@ class ScriptRunner:
 
 
     def run_script(self, script, repeat_count=1, run_name=None, logger=None, socketio=None, config=None, bo_args=None,
-                   output_path="", compiled=False, current_app=None, history=None):
+                   output_path="", compiled=False, current_app=None, history=None, optimizer=None):
         global deck
         if deck is None:
             deck = global_config.deck
-        print("history", history)
+        # print("history", history)
         if self.current_app is None:
             self.current_app = current_app
         # time.sleep(1)  # Optional: may help ensure deck readiness
@@ -82,12 +96,12 @@ class ScriptRunner:
         thread = threading.Thread(
             target=self._run_with_stop_check,
             args=(script, repeat_count, run_name, logger, socketio, config, bo_args, output_path, current_app, compiled,
-                  history)
+                  history, optimizer)
         )
         thread.start()
         return thread
 
-    def exec_steps(self, script, section_name, logger, socketio, run_id, i_progress, **kwargs):
+    def exec_steps(self, script, section_name, logger, socketio, phase_id, **kwargs):
         """
         Executes a function defined in a string line by line
         :param func_str: The function as a string
@@ -110,9 +124,15 @@ class ScriptRunner:
         # Parse function body from string
         temp_connections = global_config.defined_variables
         # Prepare execution environment
-        exec_globals = {"deck": deck, "time":time}  # Add required global objects
+        exec_globals = {"deck": deck, "time":time, "pause": pause}  # Add required global objects
         # exec_globals = {"deck": deck, "time": time, "registered_workflows":registered_workflows}  # Add required global objects
         exec_globals.update(temp_connections)
+
+        # Inject all block categories
+        for category, data in BUILDING_BLOCKS.items():
+            for method_name, method in data.items():
+                exec_globals[method_name] = method["func"]
+
         exec_locals = {}  # Local execution scope
 
         # Define function arguments manually in exec_locals
@@ -124,9 +144,9 @@ class ScriptRunner:
             if self.stop_current_event.is_set():
                 logger.info(f'Stopping execution during {section_name}')
                 step = WorkflowStep(
-                    workflow_id=run_id,
-                    phase=section_name,
-                    repeat_index=i_progress,
+                    phase_id=phase_id,
+                    # phase=section_name,
+                    # repeat_index=i_progress,
                     step_index=index,
                     method_name="stop",
                     start_time=datetime.now(),
@@ -136,31 +156,42 @@ class ScriptRunner:
                 db.session.add(step)
                 break
             line = step_list[index]
-            method_name = line.strip().split("(")[0] if "(" in line else line.strip()
-            start_time = datetime.now()
+
+            method_name = line.strip()
+            # start_time = datetime.now()
+
             step = WorkflowStep(
-                workflow_id=run_id,
-                phase=section_name,
-                repeat_index=i_progress,
+                phase_id=phase_id,
+                # phase=section_name,
+                # repeat_index=i_progress,
                 step_index=index,
                 method_name=method_name,
-                start_time=start_time,
+                start_time=datetime.now(),
             )
             db.session.add(step)
-            db.session.commit()
+            db.session.flush()
+
             logger.info(f"Executing: {line}")
             socketio.emit('execution', {'section': f"{section_name}-{index}"})
-            # self._emit_progress(socketio, 100)
+
             # if line.startswith("registered_workflows"):
             #     line = line.replace("registered_workflows.", "")
             try:
                 if line.startswith("time.sleep("): # add safe sleep for time.sleep lines
-                    duration_str = line[len("time.sleep("):-1]
+                    duration_str = line.strip()[len("time.sleep("):-1]
                     duration = float(duration_str)
                     self.safe_sleep(duration)
                 else:
                     exec(line, exec_globals, exec_locals)
-                step.run_error = False
+                # step.run_error = False
+
+            except HumanInterventionRequired as e:
+                logger.warning(f"Human intervention required: {e}")
+                socketio.emit('human_intervention', {'message': str(e)})
+                # Instead of auto-resume, explicitly stay paused until user action
+                # step.run_error = False
+                self.toggle_pause()
+
             except Exception as e:
                 logger.error(f"Error during script execution: {e}")
                 socketio.emit('error', {'message': str(e)})
@@ -168,7 +199,7 @@ class ScriptRunner:
                 step.run_error = True
                 self.toggle_pause()
             step.end_time = datetime.now()
-            # db.session.add(step)
+            step.output = exec_locals
             db.session.commit()
 
             self.pause_event.wait()
@@ -176,66 +207,93 @@ class ScriptRunner:
             # todo update script during the run
             # _func_str = script.compile()
             # step_list: list = script.convert_to_lines(_func_str).get(section_name, [])
-            if not step.run_error:
-                index += 1
-            elif not self.retry:
+            if not step.run_error or not self.retry:
                 index += 1
+
         return exec_locals  # Return the 'results' variable
 
     def _run_with_stop_check(self, script: Script, repeat_count: int, run_name: str, logger, socketio, config, bo_args,
-                             output_path, current_app, compiled, history=None):
+                             output_path, current_app, compiled, history=None, optimizer=None):
         time.sleep(1)
         # _func_str = script.compile()
         # step_list_dict: dict = script.convert_to_lines(_func_str)
         self._emit_progress(socketio, 1)
-
-        # Run "prep" section once
-        script_dict = script.script_dict
+        filename = None
+        error_flag = False
+        # create a new run entry in the database
+        repeat_mode = "batch" if config else "optimizer" if bo_args or optimizer else "repeat"
         with current_app.app_context():
-
-            run = WorkflowRun(name=script.name or "untitled", platform=script.deck or "deck",start_time=datetime.now())
+            run = WorkflowRun(name=script.name or "untitled", platform=script.deck or "deck", start_time=datetime.now(),
+                              repeat_mode=repeat_mode
+                              )
             db.session.add(run)
-            db.session.commit()
+            db.session.flush()
             run_id = run.id  # Save the ID
-            global_config.runner_status = {"id":run_id, "type": "workflow"}
-            self._run_actions(script, section_name="prep", logger=logger, socketio=socketio, run_id=run_id)
-            output_list = []
-            _, arg_type = script.config("script")
-            _, return_list = script.config_return()
+            try:
 
-            # Run "script" section multiple times
-            if repeat_count:
-                self._run_repeat_section(repeat_count, arg_type, bo_args, output_list, script,
-                                         run_name, return_list, compiled, logger, socketio,
-                                         history, output_path, run_id=run_id)
-            elif config:
-                self._run_config_section(config, arg_type, output_list, script, run_name, logger,
-                                         socketio, run_id=run_id, compiled=compiled)
+                global_config.runner_status = {"id":run_id, "type": "workflow"}
+                # Run "prep" section once
+                self._run_actions(script, section_name="prep", logger=logger, socketio=socketio, run_id=run_id)
+                output_list = []
+                _, arg_type = script.config("script")
+                _, return_list = script.config_return()
+                # Run "script" section multiple times
+                if repeat_count:
+                    self._run_repeat_section(repeat_count, arg_type, bo_args, output_list, script,
+                                             run_name, return_list, compiled, logger, socketio,
+                                             history, output_path, run_id=run_id, optimizer=optimizer)
+                elif config:
+                    self._run_config_section(config, arg_type, output_list, script, run_name, logger,
+                                             socketio, run_id=run_id, compiled=compiled)
+                # Run "cleanup" section once
+                self._run_actions(script, section_name="cleanup", logger=logger, socketio=socketio,run_id=run_id)
+                # Reset the running flag when done
+                # Save results if necessary
+                if not script.python_script and output_list:
+                    filename = self._save_results(run_name, arg_type, return_list, output_list, logger, output_path)
+                self._emit_progress(socketio, 100)
 
-            # Run "cleanup" section once
-            self._run_actions(script, section_name="cleanup", logger=logger, socketio=socketio,run_id=run_id)
-            # Reset the running flag when done
-            self.lock.release()
-            # Save results if necessary
-            filename = None
-            if not script.python_script and output_list:
-                filename = self._save_results(run_name, arg_type, return_list, output_list, logger, output_path)
-            self._emit_progress(socketio, 100)
+            except Exception as e:
+                logger.error(f"Error during script execution: {e.__str__()}")
+                error_flag = True
+            finally:
+                self.lock.release()
         with current_app.app_context():
-            run = db.session.get(WorkflowRun, run_id)  # SQLAlchemy 1.4+ recommended method
+            run = db.session.get(WorkflowRun, run_id)
             run.end_time = datetime.now()
             run.data_path = filename
+            run.run_error = error_flag
             db.session.commit()
 
+
     def _run_actions(self, script, section_name="", logger=None, socketio=None, run_id=None):
         _func_str = script.python_script or script.compile()
         step_list: list = script.convert_to_lines(_func_str).get(section_name, [])
-        logger.info(f'Executing {section_name} steps') if step_list else logger.info(f'No {section_name} steps')
+        if not step_list:
+            logger.info(f'No {section_name} steps')
+            return None
+
+        logger.info(f'Executing {section_name} steps')
         if self.stop_pending_event.is_set():
             logger.info(f"Stopping execution during {section_name} section.")
-            return
-        if step_list:
-            self.exec_steps(script, section_name, logger, socketio, run_id=run_id, i_progress=0)
+            return None
+
+        phase = WorkflowPhase(
+            run_id=run_id,
+            name=section_name,
+            repeat_index=0,
+            start_time=datetime.now()
+        )
+        db.session.add(phase)
+        db.session.flush()
+        phase_id = phase.id
+
+        step_outputs = self.exec_steps(script, section_name, logger, socketio, phase_id=phase_id)
+        # Save phase-level output
+        phase.outputs = step_outputs
+        phase.end_time = datetime.now()
+        db.session.commit()
+        return step_outputs
 
     def _run_config_section(self, config, arg_type, output_list, script, run_name, logger, socketio, run_id, compiled=True):
         if not compiled:
@@ -254,17 +312,32 @@ class ScriptRunner:
                     logger.info(f'Stopping execution during {run_name}: {i + 1}/{len(config)}')
                     break
                 logger.info(f'Executing {i + 1} of {len(config)} with kwargs = {kwargs}')
-                progress = (i + 1) * 100 / len(config)
+                progress = ((i + 1) * 100 / len(config)) - 0.1
                 self._emit_progress(socketio, progress)
                 # fname = f"{run_name}_script"
                 # function = self.globals_dict[fname]
-                output = self.exec_steps(script, "script", logger, socketio, run_id, i, **kwargs)
+
+                phase = WorkflowPhase(
+                    run_id=run_id,
+                    name="main",
+                    repeat_index=i,
+                    parameters=kwargs,
+                    start_time=datetime.now()
+                )
+                db.session.add(phase)
+                db.session.flush()
+
+                phase_id = phase.id
+                output = self.exec_steps(script, "script", logger, socketio, phase_id, **kwargs)
                 if output:
                     # kwargs.update(output)
                     output_list.append(output)
+                    phase.outputs = {k:v for k, v in output.items() if k not in arg_type.keys()}
+                phase.end_time = datetime.now()
+                db.session.commit()
 
     def _run_repeat_section(self, repeat_count, arg_types, bo_args, output_list, script, run_name, return_list, compiled,
-                            logger, socketio, history, output_path, run_id):
+                            logger, socketio, history, output_path, run_id, optimizer=None):
         if bo_args:
             logger.info('Initializing optimizer...')
             if compiled:
@@ -283,6 +356,14 @@ class ScriptRunner:
                         output_list.append(row)
                 else:
                     ax_client = bo_campaign.ax_init_form(bo_args, arg_types)
+        elif optimizer and history:
+            import pandas as pd
+            file_path = os.path.join(output_path, history)
+
+            previous_runs = pd.read_csv(file_path)
+            optimizer.append_existing_data(previous_runs)
+            for row in previous_runs:
+                output_list.append(row)
 
 
 
@@ -290,6 +371,17 @@ class ScriptRunner:
             if self.stop_pending_event.is_set():
                 logger.info(f'Stopping execution during {run_name}: {i_progress + 1}/{int(repeat_count)}')
                 break
+
+            phase = WorkflowPhase(
+                run_id=run_id,
+                name="main",
+                repeat_index=i_progress,
+                start_time=datetime.now()
+            )
+            db.session.add(phase)
+            db.session.flush()
+            phase_id = phase.id
+
             logger.info(f'Executing {run_name} experiment: {i_progress + 1}/{int(repeat_count)}')
             progress = (i_progress + 1) * 100 / int(repeat_count) - 0.1
             self._emit_progress(socketio, progress)
@@ -299,7 +391,9 @@ class ScriptRunner:
                     logger.info(f'Output value: {parameters}')
                     # fname = f"{run_name}_script"
                     # function = self.globals_dict[fname]
-                    output = self.exec_steps(script, "script", logger, socketio, run_id, i_progress, **parameters)
+                    phase.parameters = parameters
+
+                    output = self.exec_steps(script, "script", logger, socketio, phase_id, **parameters)
 
                     _output = {key: value for key, value in output.items() if key in return_list}
                     ax_client.complete_trial(trial_index=trial_index, raw_data=_output)
@@ -307,20 +401,35 @@ class ScriptRunner:
                 except Exception as e:
                     logger.info(f'Optimization error: {e}')
                     break
+            elif optimizer:
+                try:
+                    parameters = optimizer.suggest(1)
+                    logger.info(f'Output value: {parameters}')
+                    phase.parameters = parameters
+                    output = self.exec_steps(script, "script", logger, socketio, phase_id, **parameters)
+                    if output:
+                        optimizer.observe(output)
+                        output.update(parameters)
+                except Exception as e:
+                    logger.info(f'Optimization error: {e}')
+                    break
             else:
                 # fname = f"{run_name}_script"
                 # function = self.globals_dict[fname]
-                output = self.exec_steps(script, "script", logger, socketio, run_id, i_progress)
+                output = self.exec_steps(script, "script", logger, socketio, phase_id)
 
             if output:
                 output_list.append(output)
                 logger.info(f'Output value: {output}')
+                phase.outputs = output
+            phase.end_time = datetime.now()
+            db.session.commit()
 
         if bo_args:
             ax_client.save_to_json_file(os.path.join(output_path, f"{run_name}_ax_client.json"))
-        logger.info(
-            f'Optimization complete. Results saved to {os.path.join(output_path, f"{run_name}_ax_client.json")}'
-        )
+            logger.info(
+                f'Optimization complete. Results saved to {os.path.join(output_path, f"{run_name}_ax_client.json")}'
+            )
         return output_list
 
     @staticmethod
